from ravioli.complexity import calculate_complexity


def names_of(results):
    return [result.name for result in results]


# Test that functions names are extracted correctly.
def test_a_function_can_be_parsed():
    code = """
            int a_function(){}
            """
    results = calculate_complexity(code)
    assert ('a_function' in names_of(results))


def test_two_functions_can_be_parsed():
    code = """
            int a_function(){}
            int another_function(){}
            """
    results = calculate_complexity(code)
    assert ('a_function' in names_of(results))
    assert ('another_function' in names_of(results))


def test_a_commented_function_is_not_extracted():
    code = """
            //int a_function(){}
            """
    results = calculate_complexity(code)
    assert ('a_function' not in names_of(results))


def test_a_function_in_a_block_is_not_extracted():
    code = """
            /*
            int a_function(){
                //do something
            }
            */
            """
    results = calculate_complexity(code)
    assert ('a_function' not in names_of(results))


def test_a_function_with_arguments():
    code = """
            int a_function(int a){}
            """
    results = calculate_complexity(code)
    assert ('a_function' in names_of(results))


def test_a_function_with_some_whitespace():
    code = """
            int a_function (int a){}
            """
    results = calculate_complexity(code)
    assert ('a_function' in names_of(results))


def test_a_function_with_an_if():
    code = """
            int a_function (int a){
                if (a > 5) {
                    return 0;
                }
                return 1;
            }
            """
    results = calculate_complexity(code)
    assert ('a_function' in names_of(results))
    assert (len(results) == 1)


def test_a_function_with_a_while():
    code = """
            int a_function (int a){
                int x = 0;
                while (x < 5) {
                    x++;
                }
                return 1;
            }
            """
    results = calculate_complexity(code)
    assert ('a_function' in names_of(results))
    assert (len(results) == 1)


def test_a_function_with_a_for_loop():
    code = """
            int a_function (int a){
                for (int x = 0; x < 5; x++) {
                    a++;
                }
                return 1;
            }
            """
    results = calculate_complexity(code)
    assert ('a_function' in names_of(results))
    assert (len(results) == 1)


def test_a_function_with_a_do_while():
    code = """
            int a_function (int a){
                int x = 0;
                do {
                    x++;
                } while (x < 5);
                return 1;
            }
            """
    results = calculate_complexity(code)
    assert ('a_function' in names_of(results))
    assert (len(results) == 1)


def test_a_function_with_a_switch():
    code = """
            int a_function (int a){
                switch (a)
                {
                case 0:
                    return 0;
                    break;
                case 1:
                    return 1;
                    break;
                default
                    return a;
                    break;
                }
                return 1;
            }
            """
    results = calculate_complexity(code)
    assert ('a_function' in names_of(results))
    assert (len(results) == 1)


def test_a_function_which_calls_another_function():
    code = """
            int a_function (int a){
                if (a > 5) {
                    call_another_function(a + 1);
                }
                return 1;
            }
            """
    results = calculate_complexity(code)
    assert ('a_function' in names_of(results))
    assert (len(results) == 1)


def test_a_function_with_different_brace_placement():
    code = """
            int a_function (int a)
            {
                if (a > 5) {
                    call_another_function(a + 1);
                }
                return 1;
            }
            """
    results = calculate_complexity(code)
    assert ('a_function' in names_of(results))
    assert (len(results) == 1)


def test_an_entire_file_finds_all_functions():
    with open('c/sample.c') as f:
        results = calculate_complexity(f.read())
    assert(len(results) == 20)


# Test that complexity is calculated correctly.
def test_a_function_with_no_decisions_has_complexity_1():
    code = """
        int no_decisions () {
            int a_local_variable = 0;
            return 0;
        }
        """
    results = calculate_complexity(code)
    assert (results[0].complexity == 1)


def test_a_function_with_1_decision_has_complexity_2():
    code = """
            int if_else(int i) {
                if (i >= 0) {
                    return i + 1;
                }
                else {
                    return i - 1;
                }
            }
            """
    results = calculate_complexity(code)
    assert (results[0].complexity == 2)


def test_a_function_with_2_decisions_has_complexity_3():
    code = """
            int else_if(int i) {
                if (i >= 5) {
                    return i + 1;
                }
                else if (i >= 0) {
                    return i + 2;
                }
                else {
                    return i - 1;
                }
            }
            """
    results = calculate_complexity(code)
    assert (results[0].complexity == 3)


def test_a_function_with_a_nested_if():
    code = """
            int nested_if(int i) {
                if (i >= 0) {
                    if (i >= 5) {
                        return i + 2;
                    }
                    else {
                        return i + 1;
                    }
                }
                else {
                    return i - 1;
                }
            }
            """
    results = calculate_complexity(code)
    assert (results[0].complexity == 3)


def test_a_function_with_a_nested_else_if():
    code = """
            int nested_else_if(int i) {
                if (i >= 0) {
                    if (i >= 5) {
                        return i + 2;
                    }
                    else if (i >= 3) {
                        return i + 4;
                    }
                    else {
                        return i + 1;
                    }
                }
                else {
                    return i - 1;
                }
            }
            """
    results = calculate_complexity(code)
    assert (results[0].complexity == 4)


def test_a_function_with_a_nested_else_if_and_extra_statements():
    code = """
            int nested_else_if_with_extra_statements(int i) {
                i = i + 1;
                if (i >= 0) {
                    i = i + 1;
                    if (i >= 5) {
                        i = i + 1;
                        return i + 2;
                    }
                    else if (i >= 3) {
                        i = i + 1;
                        return i + 4;
                    }
                    else {
                        return i + 1;
                    }
                }
                else {
                    return i - 1;
                }
            }
            """
    results = calculate_complexity(code)
    assert (results[0].complexity == 4)


def test_a_function_with_a_while_loop():
    code = """
            int while_loop() {
                i = 0;
                while (i < 10){
                    global_variable += 1;
                    i++;
                }
            }
            """
    results = calculate_complexity(code)
    assert (results[0].complexity == 2)


def test_complexity_of_a_for_loop():
    code = """
            int for_loop() {
                for (i = 0; i < 10; i++){
                    global_variable += 1;
                }
            }
            """
    results = calculate_complexity(code)
    assert (results[0].complexity == 2)


def test_a_function_with_a_do_while_loop():
    code = """
            int do_while_loop() {
                i = 0;
                do {
                    global_variable += 1;
                    i++;
                } while (i < 10);
            }
            """
    results = calculate_complexity(code)
    assert (results[0].complexity == 2)


def test_a_function_with_a_do_while_loop_and_no_whitespace():
    code = """
            int do_while_loop() {
                i = 0;
                do {
                    global_variable += 1;
                    i++;
                }while (i < 10);
            }
            """
    results = calculate_complexity(code)
    assert (results[0].complexity == 2)


def test_complexity_of_a_switch():
    code = """
            void switch_statement(int x) {
                switch (x) {
                    case 1:
                        global_variable += 2;
                    case 2:
                        global_variable += 3;
                    default:
                        global_variable += x;
                }
            }
            """
    results = calculate_complexity(code)
    assert (results[0].complexity == 3)


def test_switch_statement_with_a_nested_if():
    code = """
            void switch_statement_with_nested_if(int x, int y) {
                switch (x) {
                    case 1:
                        if (y > 1) {
                            global_variable += 1;
                        }
                        else {
                            global_variable += 2;
                        }
                    case 2:
                        global_variable += 3;
                    default:
                        global_variable += x;
                }
            }"""
    results = calculate_complexity(code)
    assert (results[0].complexity == 4)


def test_nested_switch_statements():
    code = """
            void nested_switches(int x, int y) {
                switch (x) {
                    case 1:
                        switch(y) {
                            case 1:
                                global_variable += 5;
                                break;
                            case 2:
                                global_variable += 10;
                                break;
                            default:
                                global_variable += y;
                        }
                    case 2:
                        global_variable += 3;
                    default:
                        global_variable += x;
                }
            }"""
    results = calculate_complexity(code)
    assert (results[0].complexity == 5)


# These test for "strict cyclomatic complexity" (SCC), also called CC2.
def test_compound_conditional_in_if():
    code = """
            int compound_if(int i) {
                if ((i >= 0) && (i < 10)){
                    return i + 2;
                }
                else {
                    return i + 1;
                }
            }"""
    results = calculate_complexity(code)
    assert (results[0].complexity == 3)


def test_compound_conditional_in_if_no_whitespace():
    code = """
            int compound_if(int i) {
                if ((i >= 0)&&(i < 10)){
                    return i + 2;
                }
                else {
                    return i + 1;
                }
            }"""
    results = calculate_complexity(code)
    assert (results[0].complexity == 3)


def test_compound_conditional_with_or_in_if():
    code = """
            int compound_if(int i) {
                if ((i >= 0) || (i < 10)){
                    return i + 2;
                }
                else {
                    return i + 1;
                }
            }"""
    results = calculate_complexity(code)
    assert (results[0].complexity == 3)


def test_a_while_loop_with_a_compound_conditional():
    code = """
            int compound_while_loop(int j, int i) {
                i = 0;
                while ((j == 1) && (i < 10)) {
                    global_variable += 1;
                    i++;
                    j++;
                }
            }"""
    results = calculate_complexity(code)
    assert (results[0].complexity == 3)


def test_a_for_loop_with_a_compound_conditional():
    code = """
            int compound_for_loop() {
                for (int i = 0, j = 0; (i < 10) && (result == 0); i++) {
                    global_variable++;
                }
            }"""
    results = calculate_complexity(code)
    assert (results[0].complexity == 3)


def test_nested_for_loops():
    code = """
            int nested_for_loops() {
                for (i = 0; i < 10; i++){
                    for (j = 0; j < 10; j++) {
                        global_variable += 1;
                    }
                }
            }"""
    results = calculate_complexity(code)
    assert (results[0].complexity == 3)


def test_a_complicated_example():
    code = """
            int a_complicated_example(int x, int y, int z) {
                int answer = 0;
                if (x > 15) {
                    answer += 3;
                }
                else if ((x > 10) && (y > 5)) {
                    answer += 4;
                }
                else if ((x > 5) && (y > 3) && (z > 1)) {
                    answer += 1;
                    result = 0;
                    for (int i = 0, j = 0; (i < 10) && (result == 0); i++) {
                        if ((global_variable == 15) || (another_global == 12)) {
                            result = 1;
                        }
                    }
                }
                else {
                    answer += 10;
                }
                return answer;
            }"""
    results = calculate_complexity(code)
    assert (results[0].complexity == 11)


def test_a_more_complicated_example():
    code = """
            int a_more_complicated_example(int x, int y, int z) {
                int answer = 0;
                if (x > 15) {
                    answer += 3;
                } else if ((x > 10) && (y > 5)) {
                    answer += 4;
                } else if ((x > 5) && (y > 3) && (z > 1)) {
                    answer += 1;
                    result = 0;
                    for (int i = 0, j = 0; (i < 10) && (result == 0); i++) {
                        if ((global_variable == 15) || (another_global == 12)) {
                            result = 1;
                        }
                    }
                } else {
                    answer += 1;
                    switch (z) {
                        case 1:
                            answer += 4;
                            if (y == 2) {
                                answer += 2;
                            }
                            break;
                        case 2:
                            answer += 5;
                            break;
                        default:
                            answer += 1;
                    }
                    answer += 10;
                }
                return answer;
            }"""
    results = calculate_complexity(code)
    assert (results[0].complexity == 14)


# Get the line numbers of functions.
def test_line_number():
    code = """void a_function(){
            }
    """
    results = calculate_complexity(code)
    assert (results[0].line_number == 1)


def test_a_different_line_number():
    code = """void a_function(){
            }
            void another_function(){
            }
    """
    results = calculate_complexity(code)
    assert (results[1].line_number == 3)


def test_line_number_for_a_function_with_arguments():
    code = """
            void a_function(int a){
            }
    """
    results = calculate_complexity(code)
    assert (results[0].line_number == 2)


def test_line_number_for_a_function_with_different_braces():
    code = """
            void a_function(int a)
            {
            }
    """
    results = calculate_complexity(code)
    assert (results[0].line_number == 2)


def test_line_number_is_found_for_function_not_prototype():
    code = """
            void a_function(int a);
            void a_function(int a)
            {
            }
    """
    results = calculate_complexity(code)
    assert (results[0].line_number == 3)


def test_a_function_with_signature_split_across_lines():
    code = """
            void a_function(int a, int b,
                int c)
            {
            }
    """
    results = calculate_complexity(code)
    assert (results[0].name == 'a_function')


def test_line_number_of_a_function_with_signature_split_across_lines():
    code = """
            void a_function(int a, int b,
                int c)
            {
            }
    """
    results = calculate_complexity(code)
    assert (results[0].line_number == 2)


def test_a_function_and_prototype_with_signature_split_across_lines():
    code = """
            void a_function(int a, int b,
                int c);
            void a_function(int a, int b,
                int c)
            {
            }
    """
    results = calculate_complexity(code)
    assert (results[0].line_number == 4)


def test_a_function_and_prototype_with_split_across_three_lines():
    code = """
            void a_function(int a,
                int b,
                int c);
            void a_function(int a,
                int b,
                int c)
            {
            }
    """
    results = calculate_complexity(code)
    assert (results[0].name == 'a_function')
    assert (results[0].line_number == 5)



from ravioli.global_finder import find_globals


# Extract a list of names from the list of result objects.
def extract_names(results):
    return [result.name for result in results]


def test_a_single_global():
    code = """
            int a_global;
            """
    results = find_globals(code)
    assert ('a_global' in extract_names(results))


def test_multiple_globals():
    code = """
            int a_global;
            uint8_t another_global;
            """
    results = find_globals(code)
    assert ('a_global' in extract_names(results))
    assert ('another_global' in extract_names(results))


def test_a_global_with_assignment():
    code = """
            int a_global=0;
            """
    results = find_globals(code)
    assert ('a_global' in extract_names(results))


def test_with_assignment_and_whitespace():
    code = """
            int a_global = 0;
            """
    results = find_globals(code)
    assert ('a_global' in extract_names(results))


def test_a_function_variable_is_not_found():
    code = """
            void a_function() {
                int not_a_global;
            }
            """
    results = find_globals(code)
    assert ('not_a_global' not in extract_names(results))


def test_a_deeper_function_variable_is_not_found():
    code = """
            void a_function(int x) {
                if (x > 0) {
                    // do something
                }
                else {
                    int not_a_global;
                }
            }
            """
    results = find_globals(code)
    assert ('not_a_global' not in extract_names(results))


def test_a_static_is_not_global():
    code = """
            int a_global;
            static int not_a_global;
            """
    results = find_globals(code)
    assert ('not_a_global' not in extract_names(results))


def test_typedefs():
    code = """
            typedef int my_new_type;
            my_new_type a_global;
            """
    results = find_globals(code)
    assert(len(results) == 1)
    assert('a_global' in extract_names(results))


def test_structs():
    code = """
            struct point_t {
                int x;
                int y;
            };
            struct point_t a_global;
            """
    results = find_globals(code)
    assert(len(results) == 1)
    assert('a_global' in extract_names(results))


def test_struct_globals_declared_with_definition():
    code = """
            struct point_t {
                int x;
                int y;
            } a_global;
            """
    results = find_globals(code)
    assert(len(results) == 1)
    assert('a_global' in extract_names(results))


def test_an_extern_is_not_global():
    code = """
            int a_global;
            extern int not_a_global;
            """
    results = find_globals(code)
    assert ('not_a_global' not in extract_names(results))


def test_commented_global_is_not_global():
    code = """
            int a_global;
            /*int not_a_global;*/
            """
    results = find_globals(code)
    assert ('not_a_global' not in extract_names(results))


def test_global_in_single_line_comment_not_counted():
    code = """
            int a_global;
            //int not_a_global;
            """
    results = find_globals(code)
    assert ('not_a_global' not in extract_names(results))


def test_dont_count_preprocessor_ifs():
    code = """
            #if configUSE_PREEMPTION == 0
            {
                taskYIELD();
            }
            #endif
            """
    results = find_globals(code)
    assert ('configUSE_PREEMPTION' not in extract_names(results))


def test_dont_count_preprocessor_elifs():
    code = """
            #elif configUSE_PREEMPTION == 0
            {
                taskYIELD();
            }
            #endif
            """
    results = find_globals(code)
    assert ('configUSE_PREEMPTION' not in extract_names(results))


def test_a_name_containing_extern_is_global():
    code = """
            int external_global;
            """
    results = find_globals(code)
    assert ('external_global' in extract_names(results))


def test_a_name_containing_static_is_global():
    code = """
            int static_constant;
            """
    results = find_globals(code)
    assert ('static_constant' in extract_names(results))


def test_a_global_array():
    code = """
            char global_str[] = "abc";
            """
    results = find_globals(code)
    assert ('global_str' in extract_names(results))


def test_a_global_sized_array():
    code = """
            char global_array[10];
            """
    results = find_globals(code)
    assert ('global_array' in extract_names(results))


def test_a_global_sized_array_with_assignment():
    code = """
            int global_array[3] = [1, 2, 3];
            """
    results = find_globals(code)
    assert ('global_array' in extract_names(results))


def test_dont_count_const_globals():
    # They really don't add to the complexity.
    code = """
            const char global_str[] = "abc";
            """
    results = find_globals(code)
    assert ('global_str' not in extract_names(results))


# Find line numbers of globals.
def test_a_line_number_is_found():
    code = """int a_global;
            """
    results = find_globals(code)
    assert ('a_global' in extract_names(results))
    assert (1 == results[0].line_number)


def test_another_line_number_is_found():
    code = """int a_global;
            int another_global;
            """
    results = find_globals(code)
    assert (1 == results[0].line_number)
    assert (2 == results[1].line_number)


def test_a_line_number_with_a_comment():
    code = """// This is a comment.
            int a_global;
            int another_global;
            """
    results = find_globals(code)
    assert (2 == results[0].line_number)
    assert (3 == results[1].line_number)


def test_a_line_number_with_a_blank_line():
    code = """
            int a_global;
            int another_global;
            """
    results = find_globals(code)
    assert (2 == results[0].line_number)
    assert (3 == results[1].line_number)


def test_line_number_with_a_declaration_as_part_of_a_definition():
    code = """struct point_t {
                int x;
                int y;
            } a_global;
            """
    results = find_globals(code)
    assert (4 == results[0].line_number)


def test_find_globals_after_initialized_array():
    code = """
        u8 factoryCode[4] = {0,1,1,2};
        u64 test1;
        u64 test2;
        """
    results = find_globals(code)

    assert ('factoryCode' in extract_names(results))
    assert ('test1' in extract_names(results))
    assert ('test2' in extract_names(results))


def test_find_initialized_array_after_const_array():

    code = """
        u8 array1[4] = {1,2,3,4};
        u8 array2[4] = {5,6,7,8};
        """
    results = find_globals(code)
    assert ('array1' in extract_names(results))
    assert ('array2' in extract_names(results))


def test_handle_array_of_structs_initialization():

    code = """
        u8 array1[4] = {1,2,3,4};
        data_t array_of_structs[]={
            { .name = "Peter" },
            { .name = "James" },
            { .name = "John" },
            { .name = "Mike" }
        };
        u8 array2[4] = {5,6,7,8};
        """
    results = find_globals(code)
    assert ('array1' in extract_names(results))
    assert ('array2' in extract_names(results))
    assert ('array_of_structs' in extract_names(results))


import re

from ravioli.strip_comments import strip_comments


def count(string):
    string = strip_comments(string)
    lines = string.splitlines()
    # Remove lines containing only whitespace.
    lines = [l for l in lines if not l.isspace()]
    # Remove blank lines.
    lines = [l for l in lines if len(l) > 0]
    return len(lines)


def count_file(filename):
    with open(filename, 'r') as f:
        line_count = count(f.read())
    return line_count



import io
import sys
from ravioli.ravioli import main


def run_ravioli():
    """ Run ravioli and capture stdout.
    :return: The stdout captured during execution.
    """
    original_stdout = sys.stdout
    captured_stdout = io.StringIO()
    sys.stdout = captured_stdout
    main()
    sys.stdout = original_stdout
    return captured_stdout.getvalue()


def test_results_for_files():
    sys.argv = ["ravioli", "."]
    stdout = run_ravioli()
    assert ("sample.c" in stdout)
    assert ("main.c" in stdout)


def test_f_option_includes_globals():
    sys.argv = ["ravioli", ".", "-f"]
    stdout = run_ravioli()
    assert ("sample.c:2 global_variable" in stdout)
    assert ("sample.c:3 another_global" in stdout)


def test_x_option_parses_only_c_files():
    sys.argv = ["ravioli", "-x", "c", "."]
    stdout = run_ravioli()
    assert ("sample.c" in stdout)
    assert ("main.c" in stdout)
    assert ("another_file.cpp" not in stdout)


def test_x_option_parses_only_cpp_files():
    sys.argv = ["ravioli", "-x", "cpp", "."]
    stdout = run_ravioli()
    assert ("sample.c" not in stdout)
    assert ("main.c" not in stdout)
    assert ("another_file.cpp" in stdout)


def test_x_option_with_multiple_extensions():
    sys.argv = ["ravioli", "-x", "c", "-x", "cpp", "."]
    stdout = run_ravioli()
    assert ("sample.c" in stdout)
    assert ("main.c" in stdout)
    assert ("another_file.cpp" in stdout)


def test_find_globals_after_initialized_array_in_file():
    sys.argv = ["ravioli", "c/initialized_array.c", "-f"]
    stdout = run_ravioli()
    assert ("initialized_array.c:5 factoryCode" in stdout)
    assert ("initialized_array.c:6 test1" in stdout)
    assert ("initialized_array.c:7 test2" in stdout)


import re

from ravioli.global_variable import GlobalVariable
from ravioli.strip_comments import strip_comments


def find_globals(code):
    original_code = code
    results = []
    # Remove all comments.
    code = strip_comments(code)
    # Remove #ifs, #elifs, #ifdefs
    code = __strip_preprocessor_directives(code)
    if "some_const_array" in code:
        print(code)
    # Remove anything between brackets.
    code = re.sub(r'{[^}]*}', '{}', code, flags=re.DOTALL)
    # Find all the globals.
    global_matcher = re.compile(r'([\w\t\f\v {}]+)\s+(\w+)[\d\[\]]*(?:;|\s*=)')
    for m in global_matcher.finditer(code):
        qualifiers = m.group(1)
        name = m.group(2)
        line_number = __get_line_number(m.group(), original_code)
        if ('static' not in qualifiers
                and 'typedef' not in qualifiers
                and 'extern' not in qualifiers
                and 'const' not in qualifiers):
            results.append(GlobalVariable(name, line_number))
    return results


def __get_line_number(match, code):
    if '{}' in match:
        # Handle the special case where we've matched a global struct declared as part of its
        # definition. We strip the contents out of the brackets to help with parsing, so we
        # won't be able to find this match in the original code.
        match = match.split('}')[1]
    for line_number, line in enumerate(code.splitlines(True), 1):
        if match in line:
            return line_number
    return 0


def __strip_preprocessor_directives(code):
    code_lines = code.splitlines(True)
    non_pound_defines = [line for line in code_lines if not line.lstrip().startswith("#")]
    code = "".join(non_pound_defines)
    return code





import re


# Calculate the complexity of all the the functions in a string.
from ravioli.function import Function
from ravioli.strip_comments import strip_comments


def calculate_complexity(code):
    original_code = code
    code = strip_comments(code)
    results = []
    function_matcher = re.compile(r'\s+(\w+)\s*\([\w\s,\*]*\)\s*{', re.MULTILINE)

    for m in function_matcher.finditer(code):
        name = m.group(1)
        if __is_a_function(name):
            start_of_function = m.end()
            # We just found the next function. Extract the body of the function.
            function_body = __extract_next_function_body(code[start_of_function:])
            # Find the line number.
            line_number = __find_line_number(m.group(), original_code)

            # Compute the complexity of this function.
            results.append(Function(name, __calculate_complexity_for_a_function(function_body), line_number))

    return results


# Find the line number of this function in the code.
def __find_line_number(match, code):
    match = match.replace("{", "")
    match = match.strip()
    match = match.splitlines(True)
    code_lines = code.splitlines(True)
    for line_number in range(len(code_lines)):
        for i in range(len(match)):
            if not match[i] in code_lines[line_number + i]:
                # We didn't find a match.
                break
            if i == len(match) - 1:
                # This is the last line of the match.
                if not code_lines[line_number + i].rstrip().endswith(';'):
                    return line_number + 1
    return 0


# Determine if this name is keyword that makes a decision/
def __is_a_decision(name):
    decision_keywords = ['if', 'while', 'for']
    return name in decision_keywords


# Determine if this name is a function (and not a keyword that looks like one).
def __is_a_function(name):
    keywords_that_look_like_functions = ['if', 'while', 'for', 'switch']
    return name not in keywords_that_look_like_functions


# Extract the next function body in this section of code by matching braces.
def __extract_next_function_body(code):
    i = 0
    brace_nesting = 1
    while i < len(code) and brace_nesting > 0:
        if code[i] == '{':
            brace_nesting += 1
        elif code[i] == '}':
            brace_nesting -= 1
        i += 1
    end_index = i
    return code[:end_index]


# For strict cyclomatic complexity (SCC) count all boolean operators.
def __calculate_complexity_for_a_conditional(conditional):
    complexity = 0
    complexity += conditional.count("&&")
    complexity += conditional.count("||")
    return complexity


# Calculate the complexity of a function.
def __calculate_complexity_for_a_function(body):
    complexity = 1

    # Find if, while and for statements.
    keyword_matcher = re.compile(r'[\s}]+(\w+)\s*\((.*)\)', re.MULTILINE)
    for m in keyword_matcher.finditer(body):
        name = m.group(1)
        conditional = m.group(2)
        if __is_a_decision(name):
            # Each decision increases the complexity.
            complexity += 1
            complexity += __calculate_complexity_for_a_conditional(conditional)

    # Find case statements.
    case_matcher = re.compile(r'\s+case\s+')
    complexity += len(case_matcher.findall(body))
    return complexity



from ravioli.line_counter import count, count_file


def test_single_line():
    code = "int i = 0;"
    line_count = count(code)
    assert(line_count == 1)


def test_multiple_lines():
    code = """int i = 0;
              int j = 1;"""
    line_count = count(code)
    assert(line_count == 2)


def test_dont_count_blank_lines():
    code = """int i = 0;
    
              int j = 1;"""
    line_count = count(code)
    assert(line_count == 2)


def test_dont_count_comment_lines():
    code = """int i = 0;
              // a comment
              int j = 1;"""
    line_count = count(code)
    assert(line_count == 2)


def test_count_lines_with_trailing_comments():
    code = """int i = 0;
              bool x = true; // a comment
              int j = 1;"""
    line_count = count(code)
    assert(line_count == 3)


def test_dont_count_single_line_block_comments():
    code = """int i = 0;
              /* a comment */
              int j = 1;"""
    line_count = count(code)
    assert(line_count == 2)


def test_dont_count_block_comments_spanning_two_lines():
    code = """int i = 0;
              /* a comment
              that spans two lines */
              int j = 1;"""
    line_count = count(code)
    assert(line_count == 2)


def test_dont_count_block_comments_spanning_multiple_lines():
    code = """int i = 0;
              /* a comment
              that spans
              three lines */
              int j = 1;"""
    line_count = count(code)
    assert(line_count == 2)


def test_count_lines_in_a_file():
    line_count = count_file('c/main.c')
    assert(line_count == 6)


def test_count_lines_in_another_file():
    line_count = count_file('c/sample.c')
    assert(line_count == 247)



from ravioli.strip_comments import strip_comments


def test_multiline_block_comment():
    code = """
            /* See if the handle of the queue being unregistered in actually in the
            registry. */
            for( ux = 0; ux < configQUEUE_REGISTRY_SIZE; ux++ )
            """
    expected = """
            
            for( ux = 0; ux < configQUEUE_REGISTRY_SIZE; ux++ )
            """
    stripped = strip_comments(code)
    assert(stripped == expected)


def test_dashed_comment_line():
    code = """
            /*-----------------------------------------------------------*/
            int a;
            """
    expected = """
            
            int a;
            """
    stripped = strip_comments(code)
    assert(stripped == expected)


def test_some_tricky_code():
    code = """
                /*-----------------------------------------------------------*/
            
            #if configQUEUE_REGISTRY_SIZE > 0

                static void vQueueUnregisterQueue( xQueueHandle xQueue )
                {
                unsigned portBASE_TYPE ux;

                    /* See if the handle of the queue being unregistered in actually in the
                    registry. */
                    for( ux = 0; ux < configQUEUE_REGISTRY_SIZE; ux++ )
                    {
                        if( xQueueRegistry[ ux ].xHandle == xQueue )
                        {
                            /* Set the name to NULL to show that this slot if free again. */
                            xQueueRegistry[ ux ].pcQueueName = NULL;
                            break;
                        }
                    }
                }

            #endif
            """
    expected = """
                
            
            #if configQUEUE_REGISTRY_SIZE > 0

                static void vQueueUnregisterQueue( xQueueHandle xQueue )
                {
                unsigned portBASE_TYPE ux;

                    
                    for( ux = 0; ux < configQUEUE_REGISTRY_SIZE; ux++ )
                    {
                        if( xQueueRegistry[ ux ].xHandle == xQueue )
                        {
                            
                            xQueueRegistry[ ux ].pcQueueName = NULL;
                            break;
                        }
                    }
                }

            #endif
            """

    stripped = strip_comments(code)
    assert(stripped == expected)



def strip_comments(code):
    # Remove single line comments.
    code_lines = code.splitlines(True)
    non_comment_lines = [line for line in code_lines if not line.lstrip().startswith("//")]
    code = "".join(non_comment_lines)
    # Remove block comments
    stripped = ""
    in_comment = False
    for i, c in enumerate(code):
        if i < (len(code) - 1):
            # Only check if there are enough characters left.
            if code[i] == '/' and code[i + 1] == '*':
                in_comment = True
        if not in_comment:
            stripped += code[i]
        elif code[i - 1] == '*' and code[i] == '/':
            in_comment = False
    return stripped



from setuptools import setup

setup(name='ravioli',
      version='0.3.0',
      description='Checks for spaghetti code',
      long_description='A tool for calculating simple, useful complexity metrics -- notably the Koopman Spaghetti '
                       'Factor (KSF) -- for C. This tool is designed to work especially on embedded software written '
                       'for compilers with non-standard extensions. It works without a compiler or any preprocessing '
                       'required.',
      url='https://github.com/ElectronVector/ravioli',
      author='Matt Chernosky',
      author_email='matt@electronvector.com',
      license='MIT',
      packages=['ravioli'],
      entry_points={
            'console_scripts': ['ravioli=ravioli.ravioli:main'],
      },
      zip_safe=False)



import argparse
import os
import traceback
from operator import itemgetter
from pathlib import Path

from ravioli.complexity import calculate_complexity
from ravioli.global_finder import find_globals
from ravioli.line_counter import count


def run(filename, args):
    errors, results = process_files(args, filename)

    if args.f:
        report_all_functions(results, errors, args)
    else:
        report_ksf_for_all_modules(results, errors, args)


def process_files(args, filename):
    results = []
    errors = []
    if not os.path.isdir(filename):
        # This is a single file.
        results.append(run_single_file(filename))
    else:
        # This is a directory. Run on all the files we can find.
        source_files = get_source_files(args)

        for f in source_files:
            result = run_single_file(str(f))
            if __file_result_is_valid(result):
                # Only save results that are valid.
                results.append(result)
            if type(result) is ParsingError:
                errors.append(result)
    return errors, results


def report_all_functions(results, errors, args):

    # Print globals.
    print("-------------------------------------------------------------------------------")
    print("Globals                                                                        ")
    print("-------------------------------------------------------------------------------")
    for result in results:
        for g in result['globals_vars']:
            print(result['filename'] + ':' + str(g.line_number) + ' ' + g.name)

    # Assemble all the functions into one list.
    functions = []
    for result in results:
        for f in result['functions']:
            functions.append({'filename': result['filename'], 'line_number': f.line_number, 'name': f.name,
                              'complexity': f.complexity})

    # Sort the functions by complexity.
    functions = sorted(functions, key=itemgetter('complexity'), reverse=True)

    # Only display results above a threshold.
    functions = [f for f in functions if f['complexity'] >= args.t]

    # Print functions.
    print("-------------------------------------------------------------------------------")
    print("Functions                                                            complexity")
    print("-------------------------------------------------------------------------------")
    for f in functions:
        remaining_filename = __print_wrapped_and_indented_string(f['filename'], 72)
        print(remaining_filename + ':' + str(f['line_number']))

        # Wrap long function names.
        print("    ", end='')
        remaining_function_name = __print_wrapped_and_indented_string(f['name'], 70)
        print('{name:70} {complexity:3}'.format(name=remaining_function_name, complexity=f['complexity']))

    if args.e:
        print("-------------------------------------------------------------------------------")
        print("Errors                                                                         ")
        print("-------------------------------------------------------------------------------")
        for e in errors:
            print("*** ERROR PROCESSING: " + e.filename)
            print(e.message)


def report_ksf_for_all_modules(results, errors, args):

    # Sort by spaghetti factor.
    results = sorted(results, key=itemgetter('ksf'), reverse=True)

    # Only display results above a threshold.
    results = [r for r in results if r['ksf'] >= args.t]

    print("-------------------------------------------------------------------------------")
    print("File                                         complexity   globals   lines   ksf")
    print("-------------------------------------------------------------------------------")
    for result in results:
        remaining_filename = __print_wrapped_and_indented_string(result['filename'], 50)
        if remaining_filename != result['filename']:
            # Some of the filename has already been printed.
            print("{file:46}  {complexity:3}       {globals:3}   {loc:5}   {ksf:3}".format(
                file=remaining_filename, complexity=result['max_scc'], globals=len(result['globals_vars']),
                loc=result['loc'], ksf=result['ksf']))
        else:
            print("{file:50}  {complexity:3}       {globals:3}   {loc:5}   {ksf:3}".format(
                file=remaining_filename, complexity=result['max_scc'], globals=len(result['globals_vars']),
                loc=result['loc'], ksf=result['ksf']))

    if args.e:
        print("-------------------------------------------------------------------------------")
        print("Errors                                                                         ")
        print("-------------------------------------------------------------------------------")
        for e in errors:
            print("*** ERROR PROCESSING: " + e.filename)
            print(e.message)


class ParsingError:
    def __init__(self, filename, message):
        self.filename = filename
        self.message = message


def run_single_file(filename):
    try:
        with open(filename, 'r') as f:
            contents = f.read()
            functions = calculate_complexity(contents)
            globals_vars = find_globals(contents)
            loc = count(contents)
            # Find the maximum complexity (scc) of all functions.
            max_scc = find_max_complexity(functions)
            # Calculate the spaghetti factor.
            ksf = max_scc + (5 * len(globals_vars)) + (loc // 20)
            return {'filename': filename, 'functions': functions, 'max_scc': max_scc, 'globals_vars': globals_vars,
                    'loc': loc, 'ksf': ksf}
    except:
        # There was an error parsing this file.
        return ParsingError(filename, traceback.format_exc())


def find_max_complexity(functions):
    if len(functions) > 0:
        max_scc = max([f.complexity for f in functions])
    else:
        max_scc = 0
    return max_scc


def get_source_files(args):
    source_files = list(Path(args.source).glob('**/*.c'))
    if args.x is not None:
        source_files = []
        for ext in args.x:
            source_files += list(Path(args.source).glob('**/*.' + str(ext)))
    return source_files


def main():
    parser = argparse.ArgumentParser(description='Calculate complexity metrics for C code, specifically the Koopman '
                                                 'Spaghetti Factor (KSF).')
    parser.add_argument('source', help='the source file or folder for which to calculate metrics')
    parser.add_argument('-f', action='store_true', help='output a complete list of all globals and functions sorted '
                                                        'by complexity')
    parser.add_argument('-t', default=0, type=int, metavar='threshold', help='Only display results at or above this '
                                                                             'threshold (KSF or function complexity)')
    parser.add_argument('-e', action='store_true', help='show any errors encountered processing source files')
    parser.add_argument('-x', action='append', required=False, help='File extensions to include in the analysis (-x c -x cc -x h ...). If omitted, only .c files are analyzed')

    args = parser.parse_args()
    run(args.source, args)


def __file_result_is_valid(result):
    return type(result) is not ParsingError


def __print_wrapped_and_indented_string(str, width):
    remaining_str = str
    space_for_str = width
    while len(remaining_str) > space_for_str:
        if len(remaining_str) == len(str):
            # This is the first line.
            print(remaining_str[:space_for_str])
            remaining_str = remaining_str[space_for_str:]
            space_for_str = width - 4
        else:
            print("    ", end='')
            print(remaining_str[:space_for_str])
            remaining_str = remaining_str[space_for_str:]
    if remaining_str != str:
        # We wrapped some of the string. Print an indent so that the next
        # print is indented.
        print("    ", end='')
    return remaining_str


if __name__ == "__main__":
    main()
